//---------------------------------------------------------------------------

#include <windows.h>
#include <stdio.h>
#include <conio.h>
#include <fstream>
#include <iostream>

//Includes fr PKCS#11
#include "cryptoki.h"

//Schlsselmaterial & Daten
#include "keydata.h"

#pragma hdrstop

//---------------------------------------------------------------------------

#pragma argsused
void printBytes(CK_BYTE_PTR pBytes, DWORD dLenBytes) {

	for (DWORD d = 0; d < dLenBytes; d++) {
		printf("0x%X ",(pBytes[d]));
	}
}
int main(int argc, char* argv[])
{

   ifstream f("testcert.cer", ios::binary);

   if (!f.is_open()) {
	    LPVOID lpMsgBuf;
		LPVOID lpDisplayBuf;
		DWORD dw = GetLastError();

		FormatMessage(
			FORMAT_MESSAGE_ALLOCATE_BUFFER |
			FORMAT_MESSAGE_FROM_SYSTEM |
			FORMAT_MESSAGE_IGNORE_INSERTS,
			NULL,
			dw,
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
			(LPTSTR) &lpMsgBuf,
			0, NULL );
			printf((LPTSTR)lpMsgBuf);

		exit(0);
   }

   f.seekg(0, ios::end);
   unsigned long lFileSize = f.tellg();
   f.seekg(0, ios::beg);

   unsigned char* Certdata = new unsigned char[lFileSize];
   f.read((char*)Certdata, lFileSize);
   f.close();

   unsigned char ucPWDUser[] = { 0x31, 0x32, 0x33, 0x34, 0x35, 0x36 };
   DWORD dwLenPWDUser = 6;

   CK_FUNCTION_LIST_PTR p11;
   CK_RV returnCode;
   CK_C_GetFunctionList C_GetFunctionListDyn;

   if (argc != 2) {
		printf("Bitte LibName als Argument angeben");
		exit(0);
   }
   //DLL laden
   HINSTANCE hDll = LoadLibrary(argv[1]);
   if (hDll == NULL) {
		printf("Fehler beim Laden der Bibliothek");
		exit(0);
   }

   //Funktionen holen
   C_GetFunctionListDyn = (CK_C_GetFunctionList) GetProcAddress(hDll, "C_GetFunctionList");

   //Liste der Funktionen holen
   returnCode = C_GetFunctionListDyn(&p11);
   if (returnCode != CKR_OK) {
		FreeLibrary(hDll);
		hDll = NULL;
		p11 = NULL;
		printf("\nFehler bei C_GetFunctionList (dyn)");
		exit(0);
   }

   //PKCS#11 initialisieren
   returnCode = p11->C_Initialize(NULL_PTR);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_Initialize");
		exit(0);
   } else {
        printf("\nC_Initialize erfolgreich");
   }

   //Anzahl Slots holen
   CK_ULONG ulSlotCount;
   returnCode = p11->C_GetSlotList(CK_TRUE, NULL_PTR, &ulSlotCount);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_GetSlotList");
        exit(0);
   } else {
        printf("\nC_GetSlotList erfolgreich");
   }

   //Sind Slots da?
   if (ulSlotCount == 0) {
        printf("\nKeine Slots gefunden");
        exit(0);
   } else {
		printf("\n %d Slots gefunden !",ulSlotCount);
   }

   //Slots holen
   CK_SLOT_ID_PTR pSlotId = (CK_SLOT_ID_PTR) malloc(sizeof(CK_SLOT_ID) * ulSlotCount);
   returnCode = p11->C_GetSlotList(CK_TRUE, pSlotId, &ulSlotCount);
   if (returnCode != CKR_OK) {
        printf("\nFehler bei C_GetSlotList");
        exit(0);
   } else {
        printf("\nC_GetSlotList erfolgreich, verwende SlotID : %X",pSlotId[0]);
   }

   //1. Slot verwenden !!
   //Session ffnen
   CK_SESSION_HANDLE hSession;
   returnCode = p11->C_OpenSession(pSlotId[0], (CKF_RW_SESSION | CKF_SERIAL_SESSION), "Test-InitPin", NULL_PTR, &hSession);
   //War das OK?
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_OpenSession");
        exit(0);
   } else {
		printf("\nC_OpenSession erfolgreich, Session Handle : %X",hSession);
   }

   //Einloggen ohne PinPad
   returnCode = p11->C_Login(hSession, CKU_USER, ucPWDUser, dwLenPWDUser);
   //War das OK?
   if (returnCode != CKR_OK) {
        printf("\nFehler bei C_Login");
        exit(0);
   } else {
		printf("\nC_Login erfolgreich, User: CKU_USER, PIN: '123456', Session: %X",hSession);
   }

   //////////////////////////////////////
   // Zertifikat suchen
   // Template fr X.509 Zertifikat
   CK_OBJECT_CLASS cCert      = CKO_CERTIFICATE;
   CK_OBJECT_CLASS cPrivate   = CKO_PRIVATE_KEY;
   CK_OBJECT_CLASS cPublic    = CKO_PUBLIC_KEY;
   unsigned long ulKeySize  = 1024 / 8;

   CK_ATTRIBUTE privateKey[]    = {
		{  	CKA_CLASS,				(unsigned char *) &cPrivate, 				sizeof(cPrivate)},
		{  	CKA_ID,					(unsigned char *) &ucCKA_ID,				0x14		 	},
		{  	CKA_MODULUS,			(unsigned char *) &ucCKA_MODULUS,   	 	ulKeySize 		},
		{  	CKA_PRIVATE_EXPONENT,  	(unsigned char *) &ucCKA_PRIVATE_EXPONENT,  ulKeySize  		},
		{  	CKA_PUBLIC_EXPONENT, 	(unsigned char *) &ucCKA_PUBLIC_EXPONENT, 	3	  			},
		{  	CKA_PRIME_1,    		(unsigned char *) &ucCKA_PRIME_1,  			ulKeySize/2		},
		{  	CKA_PRIME_2,  			(unsigned char *) &ucCKA_PRIME_2, 			ulKeySize/2		},
		{  	CKA_EXPONENT_1,    		(unsigned char *) &ucCKA_EXPONENT_1,  		ulKeySize/2		},
		{  	CKA_EXPONENT_2,  		(unsigned char *) &ucCKA_EXPONENT_2, 		ulKeySize/2		},
		{  	CKA_COEFFICIENT,    	(unsigned char *) &ucCKA_COEFFICIENT,  		ulKeySize/2		},
   };

   CK_ATTRIBUTE publicKey[]    = {
		{  	CKA_CLASS,				(unsigned char *) &cPublic,					sizeof(cPublic)	},
		{  	CKA_ID,					(unsigned char *) &ucCKA_ID,		  		0x14		 	},
		{  	CKA_MODULUS,			(unsigned char *) &ucCKA_MODULUS,     		ulKeySize 		},
		{  	CKA_PUBLIC_EXPONENT, 	(unsigned char *) &ucCKA_PUBLIC_EXPONENT, 	3	  			},
   };

   CK_ATTRIBUTE privateKey2[]    = {
		{  	CKA_CLASS,				(unsigned char *) &cPrivate, 				sizeof(cPrivate)},
		{  	CKA_ID,					(unsigned char *) &ucCKA_ID2,				0x14		 	},
		{  	CKA_MODULUS,			(unsigned char *) &ucCKA_MODULUS,   	 	ulKeySize 		},
		{  	CKA_PRIVATE_EXPONENT,  	(unsigned char *) &ucCKA_PRIVATE_EXPONENT,  ulKeySize  		},
		{  	CKA_PUBLIC_EXPONENT, 	(unsigned char *) &ucCKA_PUBLIC_EXPONENT, 	3	  			},
		{  	CKA_PRIME_1,    		(unsigned char *) &ucCKA_PRIME_1,  			ulKeySize/2		},
		{  	CKA_PRIME_2,  			(unsigned char *) &ucCKA_PRIME_2, 			ulKeySize/2		},
		{  	CKA_EXPONENT_1,    		(unsigned char *) &ucCKA_EXPONENT_1,  		ulKeySize/2		},
		{  	CKA_EXPONENT_2,  		(unsigned char *) &ucCKA_EXPONENT_2, 		ulKeySize/2		},
		{  	CKA_COEFFICIENT,    	(unsigned char *) &ucCKA_COEFFICIENT,  		ulKeySize/2		},
   };

   CK_ATTRIBUTE publicKey2[]    = {
		{  	CKA_CLASS,				(unsigned char *) &cPublic,					sizeof(cPublic)	},
		{  	CKA_ID,					(unsigned char *) &ucCKA_ID2,		  		0x14		 	},
		{  	CKA_MODULUS,			(unsigned char *) &ucCKA_MODULUS,     		ulKeySize 		},
		{  	CKA_PUBLIC_EXPONENT, 	(unsigned char *) &ucCKA_PUBLIC_EXPONENT, 	3	  			},
   };

   CK_ATTRIBUTE Certificate[]    = {
		{  	CKA_CLASS,				(unsigned char *) &cCert,					sizeof(cCert)	},
		{  	CKA_ID,					(unsigned char *) &ucCKA_ID,				0x14		 	},
		{  	CKA_VALUE,				(unsigned char *) Certdata,   				lFileSize 		},
   };

   CK_ATTRIBUTE Certificate2[]    = {
		{  	CKA_CLASS,				(unsigned char *) &cCert,					sizeof(cCert)	},
		{  	CKA_ID,					(unsigned char *) &ucCKA_ID2,				0x14		 	},
		{  	CKA_VALUE,				(unsigned char *) Certdata,   				lFileSize 		},
   };

   DWORD dPrivate;
   DWORD dPublic;
   DWORD dCert;

   returnCode = p11->C_CreateObject(hSession,privateKey,10,&dPrivate);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_CreateObject: PrivateKey");
		exit(0);
   }
   returnCode = p11->C_CreateObject(hSession,publicKey,4,&dPublic);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_CreateObject: PublicKey");
		exit(0);
   }
   returnCode = p11->C_CreateObject(hSession,Certificate,3,&dCert);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_CreateObject: Certificate");
		exit(0);
   }

   returnCode = p11->C_CreateObject(hSession,privateKey2,10,&dPrivate);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_CreateObject: PrivateKey");
		exit(0);
   }
   returnCode = p11->C_CreateObject(hSession,publicKey2,4,&dPublic);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_CreateObject: PublicKey");
		exit(0);
   }
   returnCode = p11->C_CreateObject(hSession,Certificate2,3,&dCert);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_CreateObject: Certificate");
		exit(0);
   }

   returnCode = p11->C_DestroyObject(hSession,dPrivate);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_CreateObject: PrivateKey");
		exit(0);
   }
   returnCode = p11->C_DestroyObject(hSession,dPublic);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_CreateObject: PrivateKey");
		exit(0);
   }
   returnCode = p11->C_DestroyObject(hSession,dCert);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_CreateObject: PrivateKey");
		exit(0);
   }

   returnCode = p11->C_Logout(hSession);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_Logout");
		exit(0);
   }

   returnCode = p11->C_CloseSession(hSession);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_CloseSession");
		exit(0);
   }

   returnCode = p11->C_Finalize(NULL);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_Finalize");
		exit(0);
   }



   //Warten
   printf("\nTest erfolgreich !");
   getch();

}
//---------------------------------------------------------------------------
