//---------------------------------------------------------------------------

#pragma hdrstop

#include <stdio.h>
#include <iostream.h>

#include "p11Functions.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)

void printBytes(CK_BYTE_PTR pBytes, DWORD dLenBytes) {
	for (DWORD d = 0; d < dLenBytes; d++) {
		printf("0x%X ",(pBytes[d]));
	}
}

void printSignature(CK_BYTE_PTR pBytes, DWORD dLenBytes) {
	for (DWORD d = 0; d < dLenBytes; d++) {
		if ((d % 24) == 0) printf("\n   ");
		else if ((d % 8) == 0) printf(" ");
		printf("%2X ",(pBytes[d]));
	}
}

CK_FUNCTION_LIST_PTR LoadP11(char* LibName) {
   HINSTANCE hDll;
   CK_RV returnCode;
   CK_FUNCTION_LIST_PTR p11;
   CK_C_GetFunctionList C_GetFunctionListDyn;
   //DLL laden
   hDll = LoadLibrary(LibName);
   if (hDll == NULL) {
		printf("Fehler beim Laden der Bibliothek");
		return NULL;
   }

   //Funktionen holen
   C_GetFunctionListDyn = (CK_C_GetFunctionList) GetProcAddress(hDll, "C_GetFunctionList");

   //Liste der Funktionen holen
   returnCode = C_GetFunctionListDyn(&p11);
   if (returnCode != CKR_OK) {
		FreeLibrary(hDll);
		hDll = NULL;
		p11 = NULL;
		printf("\nFehler bei C_GetFunctionList (dyn)");
		return NULL;
   }

   //PKCS#11 initialisieren
   returnCode = p11->C_Initialize(NULL_PTR);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_Initialize");
		return NULL;
   }
   return p11;
}

CK_SLOT_ID selectSlot(CK_FUNCTION_LIST_PTR p11) {
	   //Anzahl Slots holen
	   CK_ULONG ulSlotCount;
	   CK_RV returnCode;
	   returnCode = p11->C_GetSlotList(CK_TRUE, NULL_PTR, &ulSlotCount);
	   if ((returnCode != CKR_OK) || (ulSlotCount == 0)) {
			printf("\nFehler bei C_GetSlotList");
			return 0;
	   }
	   CK_SLOT_ID_PTR pSlotId = (CK_SLOT_ID_PTR) malloc(sizeof(CK_SLOT_ID) * ulSlotCount);
	   returnCode = p11->C_GetSlotList(CK_TRUE, pSlotId, &ulSlotCount);
	   if (returnCode != CKR_OK) {
			printf("\nFehler bei C_GetSlotList");
			return 0;
	   }
	   printf("\n\n ... %d Slot(s) gefunden :", ulSlotCount);
	   for (CK_ULONG i = 0; i < ulSlotCount; i++) {
			printf("\n\tSlot %d: %d", i, pSlotId[i]);
	   }
	   printf("\n\n ... Slot auswaehlen (default: 0): ");
	   int i = 0;
	   std::cin >> i;
	   int SlotID = pSlotId[i];
	   free(pSlotId);
	   return SlotID;
}

CK_SESSION_HANDLE startLogInSession(CK_FUNCTION_LIST_PTR p11, CK_SLOT_ID SlotID) {
   //Session ffnen
   CK_SESSION_HANDLE hSession;
   unsigned char ucTemp[128];
   CK_RV returnCode = p11->C_OpenSession(SlotID, CKF_SERIAL_SESSION, "SDK-Test", NULL_PTR, &hSession);
   //War das OK?
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_OpenSession");
		return 0;
   }

   printf("\n\n ... Pin eingeben: ");
   std::cin >> ucTemp;

   //Einloggen ohne PinPad
   returnCode = p11->C_Login(hSession, CKU_USER, ucTemp, strlen(ucTemp));
   //War das OK?
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_Login");
		returnCode = p11->C_Finalize(NULL_PTR);
		return 0;
   }

   return hSession;
}

CK_OBJECT_HANDLE selectCertificate(CK_FUNCTION_LIST_PTR p11, CK_SESSION_HANDLE hSession) {

   CK_RV returnCode;
   CK_OBJECT_HANDLE pCertificateHandle[24];
   CK_ULONG ulMaxCount = 0;
   CK_OBJECT_CLASS cClass   = CKO_CERTIFICATE;
   CK_CERTIFICATE_TYPE   certType  = CKC_X_509;
   CK_ATTRIBUTE certTempl[]    = {
		{ CKA_CLASS, 			(unsigned char *) &cClass, 		sizeof(cClass) },
		{ CKA_CERTIFICATE_TYPE,	(unsigned char *) &certType,	sizeof(CK_CERTIFICATE_TYPE)	},
		{ CKA_LABEL,			NULL,							0 } };

   returnCode = p11->C_FindObjectsInit(hSession,certTempl,2);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_FindObjectsInit");
		returnCode = p11->C_Finalize(NULL_PTR);
		return 0;
   }
   returnCode = p11->C_FindObjects(hSession,pCertificateHandle,24,&ulMaxCount);
   if ( (returnCode != CKR_OK) || (ulMaxCount==0)) {
		printf("\nFehler bei C_FindObjects");
		return 0;
   }
   returnCode = p11->C_FindObjectsFinal(hSession);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_FindObjectsFinal");
		return 0;
   }
   printf("\n\n ... %d Zertifikat(e) gefunden :", ulMaxCount);
   CK_BYTE pbCKA_ID[128];
   CK_BYTE pbCKA_LABEL[128];
   for (CK_ULONG i = 0; i < ulMaxCount; i++) {
	   memset(pbCKA_LABEL, 0x00, 128);
	   CK_ATTRIBUTE paKeyCKA_ID[]   = { { CKA_ID, pbCKA_ID, 128 }, { CKA_LABEL, pbCKA_LABEL, 128 } };
	   p11->C_GetAttributeValue(hSession, pCertificateHandle[i], paKeyCKA_ID, 2);
	   if (returnCode != CKR_OK) {
			printf("\nFehler bei C_GetAttributeValue");
			return 0;
	   } else {
		   printf("\n\tZertifikat %d : %s", i, pbCKA_LABEL);
	   }
   }
   printf("\n\n ... Zertifikat auswaehlen (default: 0): ");
   CK_ULONG k = 0;
   std::cin >> k;
   if (k >= ulMaxCount) return 0;
   return pCertificateHandle[k];
}

CK_OBJECT_HANDLE selectPrivateKeyByCertificate(CK_FUNCTION_LIST_PTR p11, CK_SESSION_HANDLE hSession, CK_OBJECT_HANDLE Certificate) {

   CK_RV returnCode;
   CK_OBJECT_HANDLE pKeyHandle;
   CK_ULONG ulMaxCount = 0;
   CK_OBJECT_CLASS cClass   = CKO_PRIVATE_KEY;
   CK_KEY_TYPE cType = CKK_RSA;
   CK_BBOOL cTrue = CK_TRUE;
   CK_BYTE pbCKA_ID[128];
   DWORD dLenCKA_ID = 128;
   CK_ATTRIBUTE paKeyCKA_ID[] = {
		{ CKA_ID, 		pbCKA_ID, 					dLenCKA_ID } };

   p11->C_GetAttributeValue(hSession, Certificate, paKeyCKA_ID, 1);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_GetAttributeValue");
		return 0;
   }
   dLenCKA_ID = paKeyCKA_ID[0].ulValueLen;
   CK_ATTRIBUTE keyTempl[]    = {
		{ CKA_CLASS,	(unsigned char *) &cClass,	sizeof(cClass) },
		{ CKA_TOKEN,	(unsigned char *) &cTrue,	sizeof(cTrue) },
		{ CKA_ID,		(unsigned char *) &pbCKA_ID, dLenCKA_ID },
		{ CKA_KEY_TYPE,	(unsigned char *) &cType,	sizeof(cType) } };

   returnCode = p11->C_FindObjectsInit(hSession,keyTempl,4);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_FindObjectsInit");
		returnCode = p11->C_Finalize(NULL_PTR);
		return 0;
   }
   returnCode = p11->C_FindObjects(hSession,&pKeyHandle,1,&ulMaxCount);
   if ( (returnCode != CKR_OK) || (ulMaxCount==0)) {
		printf("\nFehler bei C_FindObjects");
		return 0;
   }
   returnCode = p11->C_FindObjectsFinal(hSession);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_FindObjectsFinal");
		return 0;
   }
   return pKeyHandle;
}

CK_RV signData(CK_FUNCTION_LIST_PTR p11, CK_SESSION_HANDLE hSession, CK_OBJECT_HANDLE hKey, unsigned char* pData, DWORD* dLenData) {

   CK_RV returnCode;
   CK_MECHANISM mechanism;
   mechanism.mechanism = CKM_SHA512_RSA_PKCS;
   mechanism.pParameter = NULL_PTR;
   mechanism.ulParameterLen = 0;
   CK_BYTE Signature[512];
   CK_ULONG lenSig=512;

   returnCode = p11->C_SignInit(hSession, &mechanism, hKey);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_SignInit");
		return returnCode;
   }
   returnCode = p11->C_Sign(hSession, pData, *dLenData,Signature, &lenSig);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_Sign");
		return returnCode;
   }
   (*dLenData) = lenSig;
   memcpy(pData, Signature, lenSig);
   return CKR_OK;
}

CK_RV cleanUp(CK_FUNCTION_LIST_PTR p11, CK_SESSION_HANDLE hSession) {

   CK_RV returnCode;
   returnCode = p11->C_CloseSession(hSession);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_CloseSession");
		return returnCode;
   }
   returnCode = p11->C_Finalize(NULL_PTR);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_Finalize");
		return returnCode;
   }
   return CKR_OK;
}

